print("import: 'PySide2'")
import PySide2

print("import: 'quantiphyse'")
import quantiphyse

print("import: 'quantiphyse.data'")
import quantiphyse.data

print("import: 'quantiphyse.gui'")
import quantiphyse.gui

print("import: 'quantiphyse.icons'")
import quantiphyse.icons

print("import: 'quantiphyse.packages'")
import quantiphyse.packages

print("import: 'quantiphyse.packages.core'")
import quantiphyse.packages.core

print("import: 'quantiphyse.packages.core.analysis'")
import quantiphyse.packages.core.analysis

print("import: 'quantiphyse.packages.core.batch_builder'")
import quantiphyse.packages.core.batch_builder

print("import: 'quantiphyse.packages.core.clustering'")
import quantiphyse.packages.core.clustering

print("import: 'quantiphyse.packages.core.compare'")
import quantiphyse.packages.core.compare

print("import: 'quantiphyse.packages.core.data'")
import quantiphyse.packages.core.data

print("import: 'quantiphyse.packages.core.experimental'")
import quantiphyse.packages.core.experimental

print("import: 'quantiphyse.packages.core.histogram'")
import quantiphyse.packages.core.histogram

print("import: 'quantiphyse.packages.core.overview'")
import quantiphyse.packages.core.overview

print("import: 'quantiphyse.packages.core.pca'")
import quantiphyse.packages.core.pca

print("import: 'quantiphyse.packages.core.radialprofile'")
import quantiphyse.packages.core.radialprofile

print("import: 'quantiphyse.packages.core.registration'")
import quantiphyse.packages.core.registration

print("import: 'quantiphyse.packages.core.roi_builder'")
import quantiphyse.packages.core.roi_builder

print("import: 'quantiphyse.packages.core.simulation'")
import quantiphyse.packages.core.simulation

print("import: 'quantiphyse.packages.core.smoothing'")
import quantiphyse.packages.core.smoothing

print("import: 'quantiphyse.processes'")
import quantiphyse.processes

print("import: 'quantiphyse.resources'")
import quantiphyse.resources

print("import: 'quantiphyse.test'")
import quantiphyse.test

print("import: 'quantiphyse.utils'")
import quantiphyse.utils

